# !/usr/bin/env perl

# cartesian_filter.pl
# date: 2013/06/25

# Discription: training and test data cartesian filter
# filtering condition:
# - Sentence length ratio > 2
# - zh_1_gram_common_ratio_threshold < 0.1
# - ja_1_gram_common_ratio_threshold < 0.3

use strict;
use utf8;
use Getopt::Long;
use lib './script';
use Kanji2SC;
use Util;

binmode STDIN, ":utf8";
binmode STDOUT, ":utf8";
binmode STDERR, ":utf8";

my (%opt); GetOptions(\%opt, 'data_path=s');

my $data_path = $opt{data_path};
my $zh_para_file = "$data_path/zh.para";
my $ja_para_file = "$data_path/ja.para";

my $SEN_LEN_RATIO_THRESHOLD = 2;
my $ZH_1_GRAM_COMMON_RATIO_THRESHOLD = 0.1;
my $JA_1_GRAM_COMMON_RATIO_THRESHOLD = 0.3;

&main();

sub main {
    my @ja_line_array;
    my @zh_line_array;
    
    # read en/ja sentences from parallel data
    Util::read_file_into_array("$ja_para_file", \@ja_line_array);
    # read zh sentences from parallel data
    Util::read_file_into_array("$zh_para_file", \@zh_line_array);

    my $zh_comp_file = "$data_path/zh.cartesian.filter";
    my $ja_comp_file = "$data_path/ja.cartesian.filter";
    open(OUT_ZH, ">:utf8", $zh_comp_file);
    open(OUT_JA, ">:utf8", $ja_comp_file);

    my $ja_line_array_size = @ja_line_array;
    my $zh_line_array_size = @zh_line_array;

    for(my $i = 0; $i < $ja_line_array_size; $i++) {
	# print "line: $i\n";
	my $ja_line = $ja_line_array[$i];

	for(my $j = 0; $j < $zh_line_array_size; $j++) {
	    my $zh_line = $zh_line_array[$j];

	    # comparable sentences, output after filtering
	    if($i != $j) {
		# a flag to identify whether filter the sentence pair or not
		my $filter_flag = 0;
		my $sen_len_ratio = &calc_sen_len_ratio($zh_line, $ja_line);
		my ($zh_common_ratio, $ja_common_ratio) = &calc_1_gram_common_ratio($zh_line, $ja_line);

		if($sen_len_ratio > $SEN_LEN_RATIO_THRESHOLD) {
		    $filter_flag = 1;
		}
		if($zh_common_ratio < $ZH_1_GRAM_COMMON_RATIO_THRESHOLD) {
		    $filter_flag = 1;
		}
		if($ja_common_ratio < $JA_1_GRAM_COMMON_RATIO_THRESHOLD) {
		    $filter_flag = 1;
		}

		# should be filterd
		if($filter_flag == 0) {
		    print OUT_JA "$ja_line\n";
		    print OUT_ZH "$zh_line\n";
		}
	    }
	}
    }
    close OUT_JA;
    close OUT_ZH;
}

sub calc_sen_len_ratio {
    my ($zh_line, $ja_line) = @_;
    # sentence length
    my @zh_word_array = split(/ /, $zh_line);
    my $zh_sen_len = @zh_word_array;

    my @ja_word_array = split(/ /, $ja_line);
    my $ja_sen_len = @ja_word_array;

    # sentence length difference and ratio
    my $sen_len_ratio = 0;
    if($ja_sen_len > $zh_sen_len) {
    	$sen_len_ratio = $ja_sen_len / $zh_sen_len;
    } else {
    	$sen_len_ratio = $zh_sen_len / $ja_sen_len;
    }

    return ($sen_len_ratio);
}

sub calc_1_gram_common_ratio {
    my ($zh_line, $ja_line) = @_;

    my @ja_cc_array = $ja_line =~ /\p{Han}/g;
    my $ja_cc_array_size = @ja_cc_array;
    my @zh_cc_array = $zh_line =~ /\p{Han}/g;
    my $zh_cc_array_size = @zh_cc_array;

    # ja common Chinese character flag array
    my @ja_cc_flag_array;
    for(my $i = 0; $i < $ja_cc_array_size; $i++) {
	$ja_cc_flag_array[$i] = 0;
    }
    # zh common Chinese character flag array
    my @zh_cc_flag_array;
    for(my $i = 0; $i < $zh_cc_array_size; $i++) {
	$zh_cc_flag_array[$i] = 0;
    }
    # check common Chinese character
    for(my $i = 0; $i < $ja_cc_array_size; $i++) {
	my $ja_cc = $ja_cc_array[$i];
	my @ja2sc_array = Kanji2SC::conv_j2c($ja_cc);
	foreach my $ja2sc (@ja2sc_array) {
	    for(my $j = 0; $j < $zh_cc_array_size; $j++) {
		my $zh_cc = $zh_cc_array[$j];
		if($ja2sc eq $zh_cc) {
		    $ja_cc_flag_array[$i] = 1;
		    $zh_cc_flag_array[$j] = 1;
		}
	    }
	}
    }
    # count common Chinese character
    my $ja_common_num = 0;
    for(my $i = 0; $i < $ja_cc_array_size; $i++) {
	if($ja_cc_flag_array[$i] == 1) {
	    $ja_common_num++;
	}
    }
    my $zh_common_num = 0;
    for(my $i = 0; $i < $zh_cc_array_size; $i++) {
	if($zh_cc_flag_array[$i] == 1) {
	    $zh_common_num++;
	}
    }
    # calculate ja common Chinese character ratio
    my $ja_common_ratio = 0;
    if($ja_cc_array_size == 0) {
	$ja_common_ratio = 0;
    } else {
	$ja_common_ratio = $ja_common_num / $ja_cc_array_size;
    }
    # calculate zh common Chinese character ratio
    my $zh_common_ratio = 0;
    if($zh_cc_array_size == 0) {
	$zh_common_ratio = 0;
    } else {
	$zh_common_ratio = $zh_common_num / $zh_cc_array_size;
    }

    return ($zh_common_ratio, $ja_common_ratio);
}
