# !/usr/bin/env perl

# make_event_file.pl
# author: Chu Chenhui(Kyoto University)
# email: chu@nlp.ist.i.kyoto-u.ac.jp
# date: 2013/07/08

# Discription: make event file for LIBSVM
# -input data file
# for training: zh.para.pos, ja.para.pos, zh.comp.pos, ja.comp.pos, align.para, align.comp, zh-ja.dic
# for test: zh.comp.pos, ja.comp.pos, align.comp, zh-ja.dic
# -output three event files
# 1. munteanu: features proposed in Munteanu+ 2006 (#15) (mode: 1)
# 2. content word feature (#4) (mode: 2)
# 3. POS-tag matching feature (#2)!! useless? 
# 4. kanji char feature (#21) (mode: 3)
# 5. not_kanji (ja: +not kana) feature (#9) (mode: 4)
# 6. NMT-char feature (#2) (mode: 5)
# 7. NMT-char-word feature (#2) (mode: 6)

use strict;
use utf8;
use Getopt::Long;
use lib './script';
use Kanji2SC;
use Util;

binmode STDIN, ":utf8";
binmode STDOUT, ":utf8";
binmode STDERR, ":utf8";

my (%opt); GetOptions(\%opt, 'data_path=s', 'event_file=s', 'dic_file=s', 'dic_ratio=f', 'mode=f', 'ble_dic=s');

my $data_path = $opt{data_path};
my $event_file = $opt{event_file};

my $ja_para_file = "$data_path/ja.para.pos";
my $zh_para_file = "$data_path/zh.para.pos";
my $ja_comp_file = "$data_path/ja.comp.pos";
my $zh_comp_file = "$data_path/zh.comp.pos";
my $align_para_file = "$data_path/align.para";
my $align_comp_file = "$data_path/align.comp";

my $inf = 1000;
my $para_zh_ja_score_char_file = "$data_path/para.zh_ja_score.char.rep$inf";
my $para_ja_zh_score_char_file = "$data_path/para.ja_zh_score.char.rep$inf";
my $comp_zh_ja_score_char_file = "$data_path/comp.zh_ja_score.char.rep$inf";
my $comp_ja_zh_score_char_file = "$data_path/comp.ja_zh_score.char.rep$inf";

my $para_zh_ja_score_word_file = "$data_path/para.zh_ja_score.word.rep$inf";
my $para_ja_zh_score_word_file = "$data_path/para.ja_zh_score.word.rep$inf";
my $comp_zh_ja_score_word_file = "$data_path/comp.zh_ja_score.word.rep$inf";
my $comp_ja_zh_score_word_file = "$data_path/comp.ja_zh_score.word.rep$inf";

#my $dic_file = "/plum2/chu/temp/zh-ja.dic.short";
my $mapping_file = "/home/chu/segmentation/data/pos_mapping.txt";

my %UnusedPOS_J = ('接頭辞' => 1, '接尾辞' => 1, '助詞' => 1, '助動詞' => 1, '判定詞' => 1, '指示詞' => 1, 
		   '特殊:句点' => 1, '特殊:読点' => 1, '特殊:空白' => 1, '名詞:形式名詞' => 1, '名詞:副詞的名詞' => 1);

my %UnusedPOS_C = ('AS' => 1, 'BA' => 1, 'CC' => 1, 'CS' => 1, 'DEC' => 1, 'DEG' => 1, 'DER' => 1, 'DEV' => 1, 'DT' => 1, 'IJ' => 1, 
		   'LB' => 1, 'LC' => 1, 'MSP' => 1, 'P' => 1, 'PN' => 1, 'PU' => 1, 'SB' => 1, 'SP' => 1, 'VC' => 1, 'VE' => 1);

my %UnusedPOS_E = ('CC' => 1, 'DT' => 1, 'EX' => 1, 'IN' => 1, 'LS' => 1, 'MD' => 1, 'PDT' => 1, 'POS' => 1,
		   'RP' => 1, 'TO' => 1, 'UH' => 1);

# my %dic_hash;
# &read_LEX($opt{dic_file}, \%dic_hash, $opt{dic_ratio});
# &read_LEX($opt{ble_dic}, \%dic_hash, $opt{dic_ratio}) if($opt{ble_dic});
# my $num = 0;
# foreach my $fr_word (keys %dic_hash) {
#     foreach my $en_word (keys %{$dic_hash{$fr_word}}) {
# 	$num++;
#     }
# }
# print STDERR "$num\n";

&main();

sub main {
    my %dic_hash;
#    Util::read_LEX("$dic_file", \%dic_hash);

    my $dic_ratio = 1;
    $dic_ratio = $opt{dic_ratio} if($opt{dic_ratio});
    &read_LEX($opt{dic_file}, \%dic_hash, $dic_ratio);
    &read_LEX($opt{ble_dic}, \%dic_hash, 1) if($opt{ble_dic});

#    print STDERR scalar(keys %dic_hash), "\n";

    my %pos_mapping_hash;
    &read_pos_mapping($mapping_file, \%pos_mapping_hash);

    # event file
    open(OUT_EVENT, ">:utf8", $event_file);

    open(ZH_PARA, "<:utf8", $zh_para_file);
    open(JA_PARA, "<:utf8", $ja_para_file);
    open(ALIGN_PARA, "<:utf8", $align_para_file);
    open(ZH_JA_SCORE_CHAR_PARA, "<:utf8", $para_zh_ja_score_char_file);
    open(JA_ZH_SCORE_CHAR_PARA, "<:utf8", $para_ja_zh_score_char_file);
    open(ZH_JA_SCORE_WORD_PARA, "<:utf8", $para_zh_ja_score_word_file);
    open(JA_ZH_SCORE_WORD_PARA, "<:utf8", $para_ja_zh_score_word_file);
    while(my $zh_line = <ZH_PARA>) {
    	my $ja_line = <JA_PARA>;
    	my $align_line = <ALIGN_PARA>;
	my $zh_ja_score_char = <ZH_JA_SCORE_CHAR_PARA>;
	my $ja_zh_score_char = <JA_ZH_SCORE_CHAR_PARA>;
	my $zh_ja_score_word = <ZH_JA_SCORE_WORD_PARA>;
	my $ja_zh_score_word = <JA_ZH_SCORE_WORD_PARA>;

    	chomp($zh_line);
    	chomp($ja_line);
    	chomp($align_line);
    	chomp($zh_ja_score_char);
    	chomp($ja_zh_score_char);
    	chomp($zh_ja_score_word);
    	chomp($ja_zh_score_word);

    	my ($feature) = &generate_feature($zh_line, $ja_line, \%dic_hash, \%pos_mapping_hash, $align_line, 
					  $zh_ja_score_char, $ja_zh_score_char, $zh_ja_score_word, $ja_zh_score_word,
					  1, $opt{mode});
    	print OUT_EVENT "$feature\n";
    }
    close ZH_PARA;
    close JA_PARA;
    close ALIGN_PARA;
    
    open(ZH_COMP, "<:utf8", $zh_comp_file);
    open(JA_COMP, "<:utf8", $ja_comp_file);
    open(ALIGN_COMP, "<:utf8", $align_comp_file);
    open(ZH_JA_SCORE_CHAR_COMP, "<:utf8", $comp_zh_ja_score_char_file);
    open(JA_ZH_SCORE_CHAR_COMP, "<:utf8", $comp_ja_zh_score_char_file);
    open(ZH_JA_SCORE_WORD_COMP, "<:utf8", $comp_zh_ja_score_word_file);
    open(JA_ZH_SCORE_WORD_COMP, "<:utf8", $comp_ja_zh_score_word_file);
    while(my $zh_line = <ZH_COMP>) {
    	my $ja_line = <JA_COMP>;
    	my $align_line = <ALIGN_COMP>;
	my $zh_ja_score_char = <ZH_JA_SCORE_CHAR_COMP>;
	my $ja_zh_score_char = <JA_ZH_SCORE_CHAR_COMP>;
	my $zh_ja_score_word = <ZH_JA_SCORE_WORD_COMP>;
	my $ja_zh_score_word = <JA_ZH_SCORE_WORD_COMP>;

    	chomp($zh_line);
    	chomp($ja_line);
    	chomp($align_line);
    	chomp($zh_ja_score_char);
    	chomp($ja_zh_score_char);
    	chomp($zh_ja_score_word);
    	chomp($ja_zh_score_word);

    	my ($feature) = &generate_feature($zh_line, $ja_line, \%dic_hash, \%pos_mapping_hash, $align_line, 
					  $zh_ja_score_char, $ja_zh_score_char, $zh_ja_score_word, $ja_zh_score_word,
					  -1, $opt{mode});
    	print OUT_EVENT "$feature\n";
    }
    close ZH_COMP;
    close JA_COMP;
    close ALIGN_COMP;

    close OUT_EVENT;
}

sub read_LEX {
    my ($file, $dic_hash, $dic_ratio) = @_;

    my $dic_size = Util::wc($file);
    my $dic_ratio_size = $dic_size * $dic_ratio;

    # print STDERR "$dic_size: $dic_size\n";
    # print STDERR "dic_ratio_size: $dic_ratio_size\n";

    my %rand_hash = Util::random($dic_size, $dic_ratio_size);

    my $line_num = 0;
    my $entry_num = 0;
    open(IN, "<:utf8", $file);
    while(<IN>) {
	chomp;
	my ($word_1, $word_2, $prob) = split(/ /, $_);
	if($rand_hash{$line_num}) {
	    $$dic_hash{$word_1}{$word_2} = $prob;
	    $entry_num++;
	}
	$line_num++;
    }
    close IN;
    # print STDERR "\# of Dictionary entry: $entry_num\n";
}

sub generate_feature {
    my ($zh_line, $ja_line, $dic_hash, $pos_mapping_hash, $align_line, 	$zh_ja_score_char, $ja_zh_score_char, 
	$zh_ja_score_word, $ja_zh_score_word, $flag, $mode) = @_;
    my $feature = '';

    # general feature
    my ($ja_sen_len, $zh_sen_len, $sen_len_diff, $sen_len_ratio, $ja_trans_ratio, $zh_trans_ratio,
	$ja_content_word_ratio, $zh_content_word_ratio, $ja_content_word_trans_ratio, $zh_content_word_trans_ratio,
	$ja_pos_mapping_ratio, $zh_pos_mapping_ratio) = &generate_general_feature($zh_line, $ja_line, $dic_hash, $pos_mapping_hash);

    # alignment feature
    my ($ja_null_align_num, $zh_null_align_num, $ja_null_align_per, $zh_null_align_per, $top1_fertility, $top2_fertility, $top3_fertility, 
	$longest_span_len, $longest_unconnected_len) = &generate_alignment_feature($align_line, $zh_sen_len, $ja_sen_len);

    my $zh_sen = &trans_zh_pos2sen($zh_line);
    my $ja_sen = &trans_ja_pos2sen($ja_line);

    # kanji feature
    my ($ja_cc_num, $zh_cc_num, $cc_ratio, $ja_cc_ratio, $zh_cc_ratio, $ja_1_gram_common_num, $zh_1_gram_common_num,
    	$ja_1_gram_common_ratio, $zh_1_gram_common_ratio, $ja_2_gram_common_num, $zh_2_gram_common_num, $ja_2_gram_common_ratio, 
    	$zh_2_gram_common_ratio, $ja_3_gram_common_num, $zh_3_gram_common_num, $ja_3_gram_common_ratio, $zh_3_gram_common_ratio,
    	$ja_4_gram_common_num, $zh_4_gram_common_num, $ja_4_gram_common_ratio, $zh_4_gram_common_ratio) =
    	    &generate_kanji_char_feature($zh_sen, $ja_sen) if($mode >= 3);

    # not kanji feature
    my ($ja_not_kanji_word_num, $zh_not_kanji_word_num, $not_kanji_word_ratio, 
    	$ja_not_kanji_word_ratio, $zh_not_kanji_word_ratio,
    	$ja_not_kanji_word_common_num, $zh_not_kanji_word_common_num,
    	$ja_not_kanji_word_common_ratio, $zh_not_kanji_word_common_ratio)
    	= &generate_not_kanji_feature($zh_sen, $ja_sen) if($mode >= 4 || $mode == 2.5);

    if($flag == 1) {
	$feature .= "+1 ";
    } else {
	$feature .= "-1 ";
    }

    $feature .= "1:$ja_sen_len 2:$zh_sen_len ";
    $feature .= "3:$sen_len_diff 4:$sen_len_ratio ";
    $feature .= "5:$ja_trans_ratio 6:$zh_trans_ratio ";
    $feature .= "7:$ja_null_align_num 8:$zh_null_align_num ";
    $feature .= "9:$ja_null_align_per 10:$zh_null_align_per ";
    $feature .= "11:$top1_fertility 12:$top2_fertility ";
    $feature .= "13:$top3_fertility ";
    $feature .= "14:$longest_span_len 15:$longest_unconnected_len ";

    # content feature
    if($mode >= 2) {
	$feature .= "16:$ja_content_word_ratio 17:$zh_content_word_ratio ";
	$feature .= "18:$ja_content_word_trans_ratio 19:$zh_content_word_trans_ratio ";
    }

#    $feature .= "20:$ja_pos_mapping_ratio 21:$zh_pos_mapping_ratio ";

    # do not use kanji feature, but use not kanji and content word feature
    if($mode == 2.5) {
	$feature .= "20:$ja_not_kanji_word_num 21:$zh_not_kanji_word_num 22:$not_kanji_word_ratio ";
	$feature .= "23:$ja_not_kanji_word_ratio 24:$zh_not_kanji_word_ratio ";
	$feature .= "25:$ja_not_kanji_word_common_num 26:$zh_not_kanji_word_common_num ";
	$feature .= "27:$ja_not_kanji_word_common_ratio 28:$zh_not_kanji_word_common_ratio ";
    }

    # kanji feature
    if($mode >= 3) {
	$feature .= "20:$ja_cc_num 21:$zh_cc_num 22:$cc_ratio ";
	$feature .= "23:$ja_cc_ratio 24:$zh_cc_ratio ";
	$feature .= "25:$ja_1_gram_common_num 26:$zh_1_gram_common_num ";
	$feature .= "27:$ja_1_gram_common_ratio 28:$zh_1_gram_common_ratio ";
	$feature .= "29:$ja_2_gram_common_num 30:$zh_2_gram_common_num ";
	$feature .= "31:$ja_2_gram_common_ratio 32:$zh_2_gram_common_ratio ";
	$feature .= "33:$ja_3_gram_common_num 34:$zh_3_gram_common_num ";
	$feature .= "35:$ja_3_gram_common_ratio 36:$zh_3_gram_common_ratio ";
	$feature .= "37:$ja_4_gram_common_num 38:$zh_4_gram_common_num ";
	$feature .= "39:$ja_4_gram_common_ratio 40:$zh_4_gram_common_ratio ";
    }

    # not kanji feature
    if($mode >= 4) {
	$feature .= "41:$ja_not_kanji_word_num 42:$zh_not_kanji_word_num 43:$not_kanji_word_ratio ";
	$feature .= "44:$ja_not_kanji_word_ratio 45:$zh_not_kanji_word_ratio ";
	$feature .= "46:$ja_not_kanji_word_common_num 47:$zh_not_kanji_word_common_num ";
	$feature .= "48:$ja_not_kanji_word_common_ratio 49:$zh_not_kanji_word_common_ratio ";
    }

    # NMT-char feature
    if($mode >= 5) {
	$feature .= "50:$zh_ja_score_char 51:$ja_zh_score_char ";
    }

    # NMT-word feature
    if($mode >= 6) {
	$feature .= "52:$zh_ja_score_word 53:$ja_zh_score_word ";
    }

    return ($feature);
}

sub generate_alignment_feature {
    my ($align_line, $zh_sen_len, $ja_sen_len) = @_;
    my @align_pair_array = split(/ /, $align_line);
    my %zh_index_hash;
    my %ja_index_hash;
    my %zh_pair_hash;
    my %ja_pair_hash;
    foreach my $align_pair (@align_pair_array) {
    	my ($zh_index, $ja_index) = split(/\-/, $align_pair);
    	$zh_index_hash{$zh_index}++;
    	$ja_index_hash{$ja_index}++;

    	my @ja_temp;
    	if(exists($zh_pair_hash{$zh_index})) {
    	    @ja_temp = @{$zh_pair_hash{$zh_index}};
    	    push(@ja_temp, $ja_index);
    	} else {
    	    push(@ja_temp, $ja_index);
    	}
    	$zh_pair_hash{$zh_index} = \@ja_temp;

    	my @zh_temp;
    	if(exists($ja_pair_hash{$ja_index})) {
    	    @zh_temp = @{$ja_pair_hash{$ja_index}};
    	    push(@zh_temp, $zh_index);
    	} else {
    	    push(@zh_temp, $zh_index);
    	}
    	$ja_pair_hash{$ja_index} = \@zh_temp;
    }

    # number and percentage of null alignments
    my $zh_align_num = (keys %zh_index_hash);
    my $ja_align_num = (keys %ja_index_hash);
    
    my $zh_word_num = $zh_sen_len;
    my $ja_word_num = $ja_sen_len;

    my $zh_null_align_num = $zh_word_num - $zh_align_num;
    my $ja_null_align_num = $ja_word_num - $ja_align_num;

    my $zh_null_align_per = int((($zh_null_align_num / $zh_word_num)*10000)/100);
    my $ja_null_align_per = int((($ja_null_align_num / $ja_word_num)*10000)/100);

    # top 3 largest fertilities
    my %fertility_hash;
    foreach my $zh_index (keys %zh_index_hash) {
    	my $fertility = $zh_index_hash{$zh_index};
    	my $key = 'zh' . $zh_index;
    	$fertility_hash{$key} = $fertility;
    }
    foreach my $ja_index (keys %ja_index_hash) {
    	my $fertility = $ja_index_hash{$ja_index};
    	my $key = 'en' . $ja_index;
    	$fertility_hash{$key} = $fertility;
    }
    my $fertility_i = 1;
    my $top1_fertility = 0;
    my $top2_fertility = 0;
    my $top3_fertility = 0;
    foreach my $index (sort {$fertility_hash{$b} <=> $fertility_hash{$a}} keys %fertility_hash) {
    	if($fertility_i > 3) {
    	    last;
    	}
    	my $fertility = $fertility_hash{$index};
    	if($fertility_i == 1) {
    	    $top1_fertility = $fertility;
    	}
    	if($fertility_i == 2) {
    	    $top2_fertility = $fertility;
    	}
    	if($fertility_i == 3) {
    	    $top3_fertility = $fertility;
    	}
    	$fertility_i++;
    }

    # longest contiguous connected span
    my %zh_substring_hash = &get_substring(\%zh_index_hash);
    my $zh_longest_span_len = &get_longest_span_len(\%zh_substring_hash, \%zh_pair_hash);

    my %ja_substring_hash = &get_substring(\%ja_index_hash);
    my $ja_longest_span_len = &get_longest_span_len(\%ja_substring_hash, \%ja_pair_hash);

    my $longest_span_len = $zh_longest_span_len;
    if($ja_longest_span_len > $zh_longest_span_len) {
    	$longest_span_len = $ja_longest_span_len;
    }

    # longest unconnected substring
    my $zh_unconnected_len = &get_unconnected_substring_len(\%zh_substring_hash, $zh_word_num);
    my $ja_unconnected_len = &get_unconnected_substring_len(\%ja_substring_hash, $ja_word_num);

    my $longest_unconnected_len = $zh_unconnected_len;
    if($ja_unconnected_len > $zh_unconnected_len) {
    	$longest_unconnected_len = $ja_unconnected_len;
    }

    return ($ja_null_align_num, $zh_null_align_num, $ja_null_align_per, $zh_null_align_per, 
	    $top1_fertility, $top2_fertility, $top3_fertility, $longest_span_len, $longest_unconnected_len);
}

sub generate_general_feature {
    my ($zh_line, $ja_line, $dic_hash, $pos_mapping_hash) = @_;

    # sentence length
    my @zh_word_array = split(/ /, $zh_line);
    my $zh_sen_len = @zh_word_array;

    my @ja_word_array = split(/ /, $ja_line);
    my $ja_sen_len = @ja_word_array;

    # sentence length difference and ratio
    my $sen_len_diff = 0;
    my $sen_len_ratio = 0;
    if($ja_sen_len > $zh_sen_len) {
    	$sen_len_diff = $ja_sen_len - $zh_sen_len;
    	$sen_len_ratio = $ja_sen_len / $zh_sen_len;
    } else {
    	$sen_len_diff = $zh_sen_len - $ja_sen_len;
    	$sen_len_ratio = $zh_sen_len / $ja_sen_len;
    }

    my @zh_content_word_array;
    &get_zh_content_word($zh_line, \@zh_content_word_array);
    my $zh_content_word_ratio = scalar(@zh_content_word_array) / $zh_sen_len;

    my @ja_content_word_array;
    &get_ja_content_word($ja_line, \@ja_content_word_array);
    my $ja_content_word_ratio = scalar(@ja_content_word_array) / $ja_sen_len;

    # foreach my $zh_content_word (@zh_content_word_array) {
    # 	print "$zh_content_word\n";
    # }
    # print "===========\n";
    # foreach my $ja_content_word (@ja_content_word_array) {
    # 	print "$ja_content_word\n";
    # }

    my ($ja_trans_ratio, $zh_trans_ratio) = &calc_trans_word_ratio(\@zh_word_array, \@ja_word_array, $dic_hash);
#    print "over!!\n";
    my ($ja_content_word_trans_ratio, $zh_content_word_trans_ratio) = &calc_trans_word_ratio(\@zh_content_word_array, \@ja_content_word_array, $dic_hash);
    my ($ja_pos_mapping_ratio, $zh_pos_mapping_ratio) = &calc_pos_mapping_ratio(\@zh_word_array, \@ja_word_array, $pos_mapping_hash);

    $sen_len_ratio = int(($sen_len_ratio*10000)/100);
    $zh_content_word_ratio = int(($zh_content_word_ratio*10000)/100);
    $ja_content_word_ratio = int(($ja_content_word_ratio*10000)/100);

    return ($ja_sen_len, $zh_sen_len, $sen_len_diff, $sen_len_ratio, $ja_trans_ratio, $zh_trans_ratio, 
	    $ja_content_word_ratio, $zh_content_word_ratio, $ja_content_word_trans_ratio, $zh_content_word_trans_ratio,
	    $ja_pos_mapping_ratio, $zh_pos_mapping_ratio);
}

sub get_substring {
    my ($index_hash_var) = @_;
    my %index_hash = %$index_hash_var;

    my @index_array;
    foreach my $index (sort {$a <=> $b} keys %index_hash) {
	push(@index_array, $index);
    }

    my %substring_hash;
    my $substring_id = 0;
    my @initial_array;
    push(@initial_array, $index_array[0]);
    $substring_hash{$substring_id} = \@initial_array;
    my $index_array_len = @index_array;
    for(my $i = 1; $i < $index_array_len; $i++) {
	my $prev_index = $index_array[$i-1];
	my $index = $index_array[$i];

	if($index == $prev_index + 1) {
	    my @substring_temp;
	    if(exists($substring_hash{$substring_id})) {
		@substring_temp = @{$substring_hash{$substring_id}};
		push(@substring_temp, $index);
	    } else {
		push(@substring_temp, $index);
	    }
	    $substring_hash{$substring_id} = \@substring_temp;
	} else {
	    $substring_id++;
	    my @substring_temp;
	    push(@substring_temp, $index);
	    $substring_hash{$substring_id} = \@substring_temp;
	}
    }
    return %substring_hash;
}

sub get_longest_span_len {
    my ($substring_hash_var, $pair_hash_var) = @_;
    my %substring_hash = %$substring_hash_var;
    my %pair_hash = %$pair_hash_var;

    my $max_span_len = 0;
    foreach my $substring_id (sort {$a <=> $b} keys %substring_hash) {
	my @substring_array = @{$substring_hash{$substring_id}};
	my $substring_len = @substring_array;

	my $min_corr_index = 1000;
	my $max_corr_index = 0;
	foreach my $index (@substring_array) {
	    my @inv_index_array;
	    if($pair_hash{$index}) {
	    	@inv_index_array = @{$pair_hash{$index}};
	    }
	    foreach my $inv_index (@inv_index_array) {
		if($inv_index < $min_corr_index) {
		    $min_corr_index = $inv_index;
		}
		if($inv_index > $max_corr_index) {
		    $max_corr_index = $inv_index;
		}
	    }
	}

	my $span_len = $substring_len + $max_corr_index - $min_corr_index + 1;
	if($span_len > $max_span_len) {
	    $max_span_len = $span_len;
	}
    }

    return $max_span_len;
}

sub get_unconnected_substring_len {
    my ($substring_hash_var, $word_num) = @_;
    my %substring_hash = %$substring_hash_var;

    my @min_index_array;
    my @max_index_array;
    foreach my $substring_id (sort {$a <=> $b} keys %substring_hash) {
	my @substring_array = @{$substring_hash{$substring_id}};
	my $substring_len = @substring_array;
	
	my $min_index = $substring_array[0];
	my $max_index = $substring_array[$substring_len-1];

	push(@min_index_array, $min_index);
	push(@max_index_array, $max_index);
    }

    my $len = @min_index_array;
    my $max_unconnected_len = $min_index_array[0];
    for(my $i = 1; $i < $len; $i++) {
	my $min_index = $min_index_array[$i];
	my $pre_max_index = $max_index_array[$i-1];

	my $unconnected_len = $min_index - $pre_max_index - 1;
	if($unconnected_len > $max_unconnected_len) {
	    $max_unconnected_len = $unconnected_len;
	}
    }
    my $last_unconnected_len = $word_num - 1 - $max_index_array[$len-1];
    if($last_unconnected_len > $max_unconnected_len) {
	$max_unconnected_len = $last_unconnected_len;
    }

    return $max_unconnected_len;
}

sub get_ja_content_word {
    my ($ja_line, $ja_content_word_array) = @_;

    my @ja_word_array = split(/ /, $ja_line);
    for my $ja_word_line (@ja_word_array) {
	my ($ja_word, $hinsi, $bunrui) = split(/\:/, $ja_word_line);
	next if defined($UnusedPOS_J{$hinsi});
	next if defined($UnusedPOS_J{$hinsi . ':' . $bunrui});

	push(@$ja_content_word_array, $ja_word);
    }
}

sub get_zh_content_word {
    my ($zh_line, $zh_content_word_array) = @_;

    my @zh_word_array = split(/ /, $zh_line);
    for my $zh_word_line (@zh_word_array) {
	my ($zh_word, $pos) = split(/\_/, $zh_word_line);
	next if defined($UnusedPOS_C{$pos});
	push(@$zh_content_word_array, $zh_word);
    }
}

sub calc_trans_word_ratio {
    my ($zh_word_array, $ja_word_array, $dic_hash) = @_;

    my $zh_sen_len = scalar(@$zh_word_array);
    my $ja_sen_len = scalar(@$ja_word_array);

    # array to score whether word have transltion
    my @zh_word_flag_array;
    my @ja_word_flag_array;

    # initialization
    for(my $i = 0; $i < $zh_sen_len; $i++) {
    	$zh_word_flag_array[$i] = 0;
    }
    for(my $i = 0; $i < $ja_sen_len; $i++) {
    	$ja_word_flag_array[$i] = 0;
    }

    for(my $zh_i = 0; $zh_i < $zh_sen_len; $zh_i++) {
    	my ($zh_word, $pos) = split(/\_/, $$zh_word_array[$zh_i]);

    	# check translation in en sentence
    	for(my $ja_i = 0; $ja_i < $ja_sen_len; $ja_i++) {
    	    my ($ja_word, $hinsi, $bunrui) = split(/\:/, $$ja_word_array[$ja_i]);

	    if($$dic_hash{$zh_word}{$ja_word}) {
#		print "$zh_word, $ja_word\n";
		$zh_word_flag_array[$zh_i] = 1;
		$ja_word_flag_array[$ja_i] = 1;
	    }
    	}
    }

    # counting
    my $zh_trans_num = 0;
    my $ja_trans_num = 0;
    for(my $i = 0; $i < $zh_sen_len; $i++) {
    	if($zh_word_flag_array[$i] == 1) {
#	    print "$$zh_word_array[$i]\n";
    	    $zh_trans_num++;
    	}
    }
#    print "===========\n";
    for(my $i = 0; $i < $ja_sen_len; $i++) {
    	if($ja_word_flag_array[$i] == 1) {
#	    print "$$ja_word_array[$i]\n";
    	    $ja_trans_num++;
    	}
    }
    
    my $zh_trans_ratio = $zh_trans_num / $zh_sen_len;
    my $ja_trans_ratio = $ja_trans_num / $ja_sen_len;

    $ja_trans_ratio = int(($ja_trans_ratio*10000)/100);
    $zh_trans_ratio = int(($zh_trans_ratio*10000)/100);

    return ($ja_trans_ratio, $zh_trans_ratio);
}

sub calc_pos_mapping_ratio {
    my ($zh_word_array, $ja_word_array, $pos_mapping_hash) = @_;

    my $zh_sen_len = scalar(@$zh_word_array);
    my $ja_sen_len = scalar(@$ja_word_array);

    # array to score whether word have transltion
    my @zh_word_flag_array;
    my @ja_word_flag_array;

    # initialization
    for(my $i = 0; $i < $zh_sen_len; $i++) {
    	$zh_word_flag_array[$i] = 0;
    }
    for(my $i = 0; $i < $ja_sen_len; $i++) {
    	$ja_word_flag_array[$i] = 0;
    }

    for(my $zh_i = 0; $zh_i < $zh_sen_len; $zh_i++) {
    	my ($zh_word, $pos) = split(/\_/, $$zh_word_array[$zh_i]);
    	# check translation in ja sentence
    	for(my $ja_i = 0; $ja_i < $ja_sen_len; $ja_i++) {
    	    my ($ja_word, $hinsi, $bunrui) = split(/\:/, $$ja_word_array[$ja_i]);

	    if(($$pos_mapping_hash{$hinsi}{$pos} || $$pos_mapping_hash{$bunrui}{$pos}) && ($ja_word_flag_array[$ja_i] == 0)) {
		$zh_word_flag_array[$zh_i] = 1;
		$ja_word_flag_array[$ja_i] = 1;
		last;
	    }
    	}
    }

    # counting
    my $zh_pos_mapping_num = 0;
    my $ja_pos_mapping_num = 0;
    for(my $i = 0; $i < $zh_sen_len; $i++) {
    	if($zh_word_flag_array[$i] == 1) {
    	    $zh_pos_mapping_num++;
    	}
    }
    for(my $i = 0; $i < $ja_sen_len; $i++) {
    	if($ja_word_flag_array[$i] == 1) {
    	    $ja_pos_mapping_num++;
    	}
    }
    
    my $zh_pos_mapping_ratio = $zh_pos_mapping_num / $zh_sen_len;
    my $ja_pos_mapping_ratio = $ja_pos_mapping_num / $ja_sen_len;

    $ja_pos_mapping_ratio = int(($ja_pos_mapping_ratio*10000)/100);
    $zh_pos_mapping_ratio = int(($zh_pos_mapping_ratio*10000)/100);

    return ($ja_pos_mapping_ratio, $zh_pos_mapping_ratio);
}

sub read_pos_mapping {
    my ($mapping_file, $pos_mapping_hash) = @_;

    open(IN_MAPPING, "<:utf8", $mapping_file);

    while(<IN_MAPPING>) {
	chomp;
	my ($pos_cn, $pos_jp_line) = split(/\t/, $_);
	if($pos_jp_line !~ /N\/A/) {
	    my @pos_jp_array = split(/\//, $pos_jp_line);
	    foreach my $pos_jp (@pos_jp_array) {
		if($pos_jp =~ /(.*?)\((.*?)\)/) {
		    my @sub_poses = split(/ /, $2);
		    foreach my $sub_pos (@sub_poses) {
			$$pos_mapping_hash{$sub_pos}{$pos_cn}++;
		    }
		} else {
		    $$pos_mapping_hash{$pos_jp}{$pos_cn}++;
		}
	    }
	}
    }
    close IN_MAPPING;
}

sub generate_kanji_char_feature {
    my ($zh_line, $ja_line) = @_;
    # Chinese character number
    my @ja_cc_array = $ja_line =~ /\p{Han}/g;
    my $ja_cc_num = @ja_cc_array;

    my @zh_cc_array = $zh_line =~ /\p{Han}/g;
    my $zh_cc_num = @zh_cc_array;

    # Chinese character ratio
    my $cc_ratio = 0;
    if($ja_cc_num > $zh_cc_num) {
    	if($zh_cc_num > 0) {
    	    $cc_ratio = $ja_cc_num / $zh_cc_num;
    	}
    } else {
    	if($ja_cc_num > 0) {
    	    $cc_ratio = $zh_cc_num / $ja_cc_num;
    	}
    }

    # Ja Chinese character ratio
    $ja_line =~ s/ //g;
    my @ja_char_array = $ja_line =~ /./g;
    my $ja_char_num = @ja_char_array;
    my $ja_cc_ratio = $ja_cc_num / $ja_char_num;
    # Zh Chinese character ratio
    $zh_line =~ s/ //g;
    my @zh_char_array = $zh_line =~ /./g;
    my $zh_char_num = @zh_char_array;
    my $zh_cc_ratio = $zh_cc_num / $zh_char_num;

    # Common Chinese character ratio: Ja & Zh
    # 1-gram feature
    my @ja_1_gram_array = &get_kanji_n_gram(\@ja_char_array, 1);
    my @zh_1_gram_array = &get_kanji_n_gram(\@zh_char_array, 1);
    my @result_array_1_gram = &get_common_kanji_n_gram_num(\@ja_1_gram_array, \@zh_1_gram_array);
    my $ja_1_gram_common_num = $result_array_1_gram[0];
    my $zh_1_gram_common_num = $result_array_1_gram[1];
    my $ja_1_gram_common_ratio = $result_array_1_gram[2];
    my $zh_1_gram_common_ratio = $result_array_1_gram[3];

    # 2-gram feature
    my @ja_2_gram_array = &get_kanji_n_gram(\@ja_char_array, 2);
    my @zh_2_gram_array = &get_kanji_n_gram(\@zh_char_array, 2);
    my @result_array_2_gram = &get_common_kanji_n_gram_num(\@ja_2_gram_array, \@zh_2_gram_array);
    my $ja_2_gram_common_num = $result_array_2_gram[0];
    my $zh_2_gram_common_num = $result_array_2_gram[1];
    my $ja_2_gram_common_ratio = $result_array_2_gram[2];
    my $zh_2_gram_common_ratio = $result_array_2_gram[3];

    # 3-gram feature
    my @ja_3_gram_array = &get_kanji_n_gram(\@ja_char_array, 3);
    my @zh_3_gram_array = &get_kanji_n_gram(\@zh_char_array, 3);
    my @result_array_3_gram = &get_common_kanji_n_gram_num(\@ja_3_gram_array, \@zh_3_gram_array);
    my $ja_3_gram_common_num = $result_array_3_gram[0];
    my $zh_3_gram_common_num = $result_array_3_gram[1];
    my $ja_3_gram_common_ratio = $result_array_3_gram[2];
    my $zh_3_gram_common_ratio = $result_array_3_gram[3];

    # 4-gram feature
    my @ja_4_gram_array = &get_kanji_n_gram(\@ja_char_array, 4);
    my @zh_4_gram_array = &get_kanji_n_gram(\@zh_char_array, 4);
    my @result_array_4_gram = &get_common_kanji_n_gram_num(\@ja_4_gram_array, \@zh_4_gram_array);
    my $ja_4_gram_common_num = $result_array_4_gram[0];
    my $zh_4_gram_common_num = $result_array_4_gram[1];
    my $ja_4_gram_common_ratio = $result_array_4_gram[2];
    my $zh_4_gram_common_ratio = $result_array_4_gram[3];

    # treat ratio as interger
    $cc_ratio = int(($cc_ratio*10000)/100);
    $ja_cc_ratio = int(($ja_cc_ratio*10000)/100);
    $zh_cc_ratio = int(($zh_cc_ratio*10000)/100);

    return ($ja_cc_num, $zh_cc_num, $cc_ratio, $ja_cc_ratio, $zh_cc_ratio, $ja_1_gram_common_num, $zh_1_gram_common_num,
	    $ja_1_gram_common_ratio, $zh_1_gram_common_ratio, $ja_2_gram_common_num, $zh_2_gram_common_num, $ja_2_gram_common_ratio, 
	    $zh_2_gram_common_ratio, $ja_3_gram_common_num, $zh_3_gram_common_num, $ja_3_gram_common_ratio, $zh_3_gram_common_ratio,
	    $ja_4_gram_common_num, $zh_4_gram_common_num, $ja_4_gram_common_ratio, $zh_4_gram_common_ratio);
}

sub get_kanji_n_gram {
    my ($char_array_var, $n) = @_;
    my @char_array = @$char_array_var;

    my @n_gram_array = ();
    my $len = @char_array;
    for(my $i = 0; $i <= $len-$n; $i++) {
	my $flag = 1;
	my $str = '';
	for(my $j = $i; $j < $i+$n; $j++) {
	    my $char = $char_array[$j];
	    $str .= $char;
	    if($char !~ /\p{Han}/) {
		$flag = 0;
		last;
	    }
	}
	if($flag == 1) {
	    push(@n_gram_array, $str);
	}
    }

    return @n_gram_array;
}

sub get_common_kanji_n_gram_num {
    my ($ja_kanji_n_gram_var, $zh_kanji_n_gram_var) = @_;
    my @ja_kanji_n_gram = @$ja_kanji_n_gram_var;
    my @zh_kanji_n_gram = @$zh_kanji_n_gram_var;

    my @result_array = ();

    my $ja_cc_num = @ja_kanji_n_gram;
    my $zh_cc_num = @zh_kanji_n_gram;

    # ja common Kanji n-gram flag array
    my @ja_cc_flag_array;
    for(my $i = 0; $i < $ja_cc_num; $i++) {
	$ja_cc_flag_array[$i] = 0;
    }
    # zh common Kanji n-gram flag array
    my @zh_cc_flag_array;
    for(my $i = 0; $i < $zh_cc_num; $i++) {
	$zh_cc_flag_array[$i] = 0;
    }
    # check common Chinese character
    for(my $i = 0; $i < $ja_cc_num; $i++) {
	my $ja_cc = $ja_kanji_n_gram[$i];
	my @ja2sc_array = Kanji2SC::conv_j2c_word($ja_cc);
	foreach my $ja2sc (@ja2sc_array) {
	    for(my $j = 0; $j < $zh_cc_num; $j++) {
		my $zh_cc = $zh_kanji_n_gram[$j];
		if($ja2sc eq $zh_cc) {
		    $ja_cc_flag_array[$i] = 1;
		    $zh_cc_flag_array[$j] = 1;
		}
	    }
	}
    }
    # count common Chinese character
    my $ja_common_num = 0;
    for(my $i = 0; $i < $ja_cc_num; $i++) {
	if($ja_cc_flag_array[$i] == 1) {
	    $ja_common_num++;
	}
    }
    my $zh_common_num = 0;
    for(my $i = 0; $i < $zh_cc_num; $i++) {
	if($zh_cc_flag_array[$i] == 1) {
	    $zh_common_num++;
	}
    }
    # calculate ja common Chinese character ratio
    my $ja_common_ratio = 0;
    if($ja_cc_num == 0) {
	$ja_common_ratio = 0;
    } else {
	$ja_common_ratio = $ja_common_num / $ja_cc_num;
    }
    # calculate zh common Chinese character ratio
    my $zh_common_ratio = 0;
    if($zh_cc_num == 0) {
	$zh_common_ratio = 0;
    } else {
	$zh_common_ratio = $zh_common_num / $zh_cc_num;
    }

    $ja_common_ratio = int(($ja_common_ratio*10000)/100);
    $zh_common_ratio = int(($zh_common_ratio*10000)/100);

    push(@result_array, $ja_common_num);
    push(@result_array, $zh_common_num);
    push(@result_array, $ja_common_ratio);
    push(@result_array, $zh_common_ratio);

    return @result_array;
}

sub trans_ja_pos2sen {
    my ($ja_line) = @_;
    my $ja_sen;
    my @ja_word_array = split(/ /, $ja_line);
    for my $ja_word_line (@ja_word_array) {
	my ($ja_word, $hinsi, $bunrui) = split(/\:/, $ja_word_line);
	$ja_sen .= $ja_word;
	$ja_sen .= ' ';
    }
    $ja_sen =~ s/\s\Z//g;
    return $ja_sen;
}

sub trans_zh_pos2sen {
    my ($zh_line) = @_;
    my $zh_sen;
    my @zh_word_array = split(/ /, $zh_line);
    for my $zh_word_line (@zh_word_array) {
	my ($zh_word, $pos) = split(/\_/, $zh_word_line);
	$zh_sen .= $zh_word;
	$zh_sen .= ' ';
    }
    $zh_sen =~ s/\s\Z//g;
    return $zh_sen;
}

sub generate_not_kanji_feature {
    my ($zh_line, $ja_line) = @_;

    # not kanji word features
    my @zh_not_kanji_word_array = &get_not_kanji_word($zh_line);
    my @ja_not_kanji_word_array = &get_not_kanji_word($ja_line);
    my $ja_not_kanji_word_num = scalar(@ja_not_kanji_word_array);
    my $zh_not_kanji_word_num = scalar(@zh_not_kanji_word_array);

    # not kanji word ratio
    my $not_kanji_word_ratio = 0;
    if($ja_not_kanji_word_num > $zh_not_kanji_word_num) {
    	if($zh_not_kanji_word_num > 0) {
    	    $not_kanji_word_ratio = $ja_not_kanji_word_num / $zh_not_kanji_word_num;
    	}
    } else {
    	if($ja_not_kanji_word_num > 0) {
    	    $not_kanji_word_ratio = $zh_not_kanji_word_num / $ja_not_kanji_word_num;
    	}
    }
    # sentence length
    my @zh_word_array = split(/ /, $zh_line);
    my $zh_sen_len = @zh_word_array;
    my @ja_word_array = split(/ /, $ja_line);
    my $ja_sen_len = @ja_word_array;
    # ja not kanji word ratio
    my $ja_not_kanji_word_ratio = $ja_not_kanji_word_num / $ja_sen_len;
    # zh not kanji word ratio
    my $zh_not_kanji_word_ratio = $zh_not_kanji_word_num / $zh_sen_len;

    my @result_array_not_kanji_word = &get_common_kanji_n_gram_num(\@ja_not_kanji_word_array, \@zh_not_kanji_word_array);
    my $ja_not_kanji_word_common_num = $result_array_not_kanji_word[0];
    my $zh_not_kanji_word_common_num = $result_array_not_kanji_word[1];
    my $ja_not_kanji_word_common_ratio = $result_array_not_kanji_word[2];
    my $zh_not_kanji_word_common_ratio = $result_array_not_kanji_word[3];

    # treat ratio as interger
    $not_kanji_word_ratio = int(($not_kanji_word_ratio*10000)/100);
    $ja_not_kanji_word_ratio = int(($ja_not_kanji_word_ratio*10000)/100);
    $zh_not_kanji_word_ratio = int(($zh_not_kanji_word_ratio*10000)/100);

    return ($ja_not_kanji_word_num, $zh_not_kanji_word_num, $not_kanji_word_ratio, 
    	    $ja_not_kanji_word_ratio, $zh_not_kanji_word_ratio,
    	    $ja_not_kanji_word_common_num, $zh_not_kanji_word_common_num,
    	    $ja_not_kanji_word_common_ratio, $zh_not_kanji_word_common_ratio);  
}

sub get_not_kanji_word {
    my ($seg_sen) = @_;
    my @word_array = split(/ /, $seg_sen);

    my @not_kanji_word_array = ();
    foreach my $word (@word_array) {
	if($word !~ /\p{Hiragana}|\p{Katakana}|\p{Han}/) {
	    push(@not_kanji_word_array, $word);
	}
    }
    return @not_kanji_word_array;
}
